package ca.ualberta.xdong.plasMapper.annotate;
import java.util.*;
import java.io.*;

/**
 * 
 * @author xdong
 *
 */
public class PlasFeatures {

	//fasta format dna sequence
	protected String sequence;

	//all the vector features
	protected Vector hitFeature;

	//for get enzyme
	protected PlasmidCleavageSiteAnnotate site;

	protected String outputFile;
	protected String inputFile;

	/**
	 * Create a new PlasFeature instance by usig input sequence.
	 * <p>
	 * it uses input sequence to do blast search against local feature databas
	 * and blast search output format is format 8. It then parse the output to
	 * get all the features of the input sequence and put it into hitFeature
	 * vector.
	 * <p>
	 * It also use biojava to do the restriction site annotation and get a
	 * cleavage site object.
	 * 
	 * @param sequence
	 *            input sequence
	 * @param inputFile
	 *            input sequence file
	 * @param outputFile
	 *            blast search output file
	 * @throws Exception
	 */

	public PlasFeatures(
		boolean file,
		String sequence,
		String uploadFile,
		String inputFile,
		String outputFile)
		throws Exception {

		this.outputFile = outputFile;
		this.inputFile = inputFile;

		if (file) {

			StringBuffer buffer = Utility.readString(uploadFile);
			this.sequence = Utility.removeUselessCharacter(new String(buffer));
		} else {
			this.sequence = Utility.removeUselessCharacter(sequence);
		}

		DataOutputStream vectorSeq =
			new DataOutputStream(new FileOutputStream(inputFile));
		vectorSeq.writeBytes(Utility.formatToFasta(this.sequence));
		vectorSeq.close();

		PlasmidFeatureAnnotate annotate =
			new PlasmidFeatureAnnotate(inputFile, outputFile);
		this.hitFeature = annotate.getHitFeatures();

		this.site = new PlasmidCleavageSiteAnnotate(this.sequence);
	}

	public PlasmidCleavageSiteAnnotate getSite() {
		return this.site;
	}

	/*
	  get the sequence of the plasmid
	*/
	public String getSequence() {
		//return site.getSeq();
		return this.sequence;

	}

	/*
	  get the sequence length
	*/
	public int getSeqLength() {

		//return site.getSeqLength();
		return this.sequence.length();
	}

	/**
	 * this function looks for all the ORF that is longer or equals to minLen
	 * the minLen is the number of the codon
	 * In this method, we look for start codon first, then look for a stop
	 * codon
	 * 
	 * @param minLen	minimum length of a open reading frame
	 * @param strands	define the direction of the strand. It will either be
	 * 0, 1, 2
	 * @return a vector of Feature sites that represents a collection 
	 * of the open reading frame
	 */

	public Vector getStartORFs(int minLen, String[] strands) {

		int length = getSeqLength();

		PlasmidTool plasmidTool = new PlasmidTool(this.sequence);
		Vector orfs = plasmidTool.getStartORFs(minLen);

		Vector result = new Vector(10);
		int counter = 0;
		for (int i = 0; i < orfs.size(); i++) {

			OpenReadingFrame orf = (OpenReadingFrame) orfs.elementAt(i);
			int frameIndex = orf.getFrameIndex();
			int strand = 0;

			//forward
			if (frameIndex <= 3)
				strand = PlasMapperConstants.FORWARD;

			//reverse
			else if (frameIndex >= 4)
				strand = PlasMapperConstants.REVERSE;

			if (strands != null && strands.length > 0) {
				//forward
				for (int j = 0; j < strands.length; j++) {

					if ((Integer.parseInt(strands[j])) == strand) {
						counter++;

						if (strand == PlasMapperConstants.FORWARD) {
							FeatureSite site =
								new FeatureSite(
									((orf.getStart()) * 3) + frameIndex,
									3 * (orf.getStop() + 1) + frameIndex - 1,
									"ORF_"
										+ counter
										+ " rf("
										+ frameIndex
										+ ")",
									strand + "");
							result.add(site);
						} else if (strand == PlasMapperConstants.REVERSE) {
							FeatureSite site =
								new FeatureSite(
									length
										- (3 * (orf.getStop()) + frameIndex - 2),
									length
										- ((orf.getStart()) * 3)
										- frameIndex
										+ 4,
									"ORF_"
										+ counter
										+ " rf("
										+ frameIndex
										+ ")",
									strand + "");
							result.add(site);
						}
					}
				}
			}
		}
		return result;

	}


	/**
	 * this function looks for all the ORF that is longer or equals to minLen
	 * the minLen is the number of the codon
	 * In this method, we just look for the sequence between stop codon
	 * 
	 * @param minLen	minimum length of a open reading frame
	 * @param strands	define the direction of the strand. It will either be
	 * 1, 2
	 * @return a vector of Feature sites that represents a collection 
	 * of the open reading frame
	 */
	
	public Vector getORFs(int minLen, String[] strands) {

		int length = getSeqLength();

		PlasmidTool plasmidTool = new PlasmidTool(this.sequence);

		Vector orfs = plasmidTool.getORFs(minLen);
		//System.out.println("orf=" + orfs.toString());

		Vector result = new Vector(10);
		int counter = 0;
		for (int i = 0; i < orfs.size(); i++) {

			OpenReadingFrame orf = (OpenReadingFrame) orfs.elementAt(i);
			int frameIndex = orf.getFrameIndex();
			int strand = 0;

			//forward
			if (frameIndex <= 3)
				strand = PlasMapperConstants.FORWARD;

			//reverse
			else if (frameIndex >= 4)
				strand = PlasMapperConstants.REVERSE;

			if (strands != null && strands.length > 0) {
				//forward
				for (int j = 0; j < strands.length; j++) {
					//System.out.println(strands[j]);
					if ((Integer.parseInt(strands[j])) == strand) {
						counter++;

						if (strand == PlasMapperConstants.FORWARD) {
							FeatureSite site =
								new FeatureSite(
									((orf.getStart()) * 3) + 1,
									(3 * (orf.getStop()) + 2)
										+ 1
										+ frameIndex
										- 1,
									"ORF_"
										+ counter
										+ " rf("
										+ frameIndex
										+ ")",
									strand + "");
							result.add(site);
						} else if (strand == PlasMapperConstants.REVERSE) {
							FeatureSite site =
								new FeatureSite(
									length
										- (3 * (orf.getStop())
											+ 2
											+ frameIndex
											- 4),
									length - ((orf.getStart()) * 3),
									"ORF_"
										+ counter
										+ " rf("
										+ frameIndex
										+ ")",
									strand + "");
							result.add(site);
						}
					}
				}
			}
		}
		return result;

	}

	/**
	 * 
	 * @param enzymeType 0 for all enzyme set, 1 for common enzyme set
	 * @return	a collection of FeatureSite objects represent unique
	 * 			Restriction sites annotated by using the given enzyme set
	 * @throws Exception
	 */
	public Vector getUniqueRestrictionSites(int enzymeType) throws Exception {

		Vector siteVector = this.site.getRestrictionSites(enzymeType);
		Vector uniqueVec = this.site.getUniqueRestrictionSites(siteVector);
		return uniqueVec;

	}
	/**
		 * 
		 * @param enzymeType 0 for all enzyme set, 1 for common enzyme set
		 * @return	a collection of FeatureSite objects represent unique
		 * 			Restriction sites annotated by using the given enzyme set
		 * @throws Exception
		 */
		public Vector getUniqueRestrictionSitesWithStartEndPos(int enzymeType) throws Exception {

			Vector siteVector = this.site.getRestrictionSitesWithStartEndPos(enzymeType);
			Vector uniqueVec = this.site.getUniqueRestrictionSites(siteVector);
			return uniqueVec;

		}

	/**
	 * 
	 * @param 	enzymetType 0 means all enzyme set, 1 means common
	 * 			enzyme set
	 * @return	a collection of FeatureSites that represent non-unique
	 * 			restriction sites by using give enzyme type
	 * @throws 	Exception
	 */
	public Vector getNotUniqueRestrictionSites(int enzymetType)
		throws Exception {

		
		Vector siteVector = this.site.getRestrictionSites(enzymetType);
		Vector uniqueVec = this.site.getUniqueRestrictionSites(siteVector);

		Vector notUnique = new Vector(10);

		for (int i = 0; i < siteVector.size(); i++) {

			FeatureSite site1 = (FeatureSite) siteVector.elementAt(i);

			int j = 0;

			for (; j < uniqueVec.size(); j++) {

				FeatureSite site2 = (FeatureSite) uniqueVec.elementAt(j);
				if ((site1.getLabel()).equals(site2.getLabel()))
					break;
			}

			if (j == uniqueVec.size())
				notUnique.add(site1);
		}
		return notUnique;

	}
	
	/**
	 * Get a collection of FeatureSite objects which represent unique restriction
	 * site annotated by using a collection of certain cutters from given enzyme set
	 * @param ncutter 
	 * @param enzymeType 0 means all enzyme set, 1 means common enzyme set
	 * @return
	 * @throws Exception
	 */
	public Vector getUniqueRestrictionSites(Vector ncutter, int enzymeType)
		throws Exception {
		Vector siteVector = this.site.getRestrictionSites(ncutter, enzymeType);
		Vector uniqueVec = this.site.getUniqueRestrictionSites(siteVector);
		return uniqueVec;

	}
	/**
		 * Get a collection of FeatureSite objects which represent unique restriction
		 * site with start and end position annotated by using a collection of certain cutters from given enzyme set
		 * @param ncutter 
		 * @param enzymeType 0 means all enzyme set, 1 means common enzyme set
		 * @return
		 * @throws Exception
		 */
		public Vector getUniqueRestrictionSitesWithStartEndPos(Vector ncutter, int enzymeType)
			throws Exception {
			Vector siteVector = this.site.getRestrictionSitesWithStartEndPos(ncutter, enzymeType);
			Vector uniqueVec = this.site.getUniqueRestrictionSites(siteVector);
			return uniqueVec;

		}

	/**
	 * Get a collection of FeatureSite objects which represent non-unique restriction
	 * site annotated by using a collection of certain cutters from given enzyme set
	 * @param ncutter 
	 * @param enzymeType 0 means all enzyme set, 1 means common enzyme set
	 * @return
	 * @throws Exception
	 */
	public Vector getNotUniqueRestrictionSites(Vector ncutter, int enzymeType)
		throws Exception {
		Vector siteVector = site.getRestrictionSites(ncutter, enzymeType);
		//Vector siteVector = getAllRestrictionSites(ncutter);
		Vector uniqueVec = site.getUniqueRestrictionSites(siteVector);
		Vector notUnique = new Vector(10);

		for (int i = 0; i < siteVector.size(); i++) {

			FeatureSite site1 = (FeatureSite) siteVector.elementAt(i);

			int j = 0;

			for (; j < uniqueVec.size(); j++) {

				FeatureSite site2 = (FeatureSite) uniqueVec.elementAt(j);
				if ((site1.getLabel()).equals(site2.getLabel()))
					break;
			}

			if (j == uniqueVec.size())
				notUnique.add(site1);
		}
		return notUnique;

	}

	/**
	 * Get a collection of FeatureSite objects which represent all the 
	 * promoter features in the sequence
	 * 
	 * @return	a collection of FeatureSite objects
	 */
	public Vector getPromoter() {

		String temp = "";
		int index = 0;
		int last = 0;
		
		Vector pro = new Vector(5);
		FeatureSite site = null;

		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();

			if (temp.indexOf("[PRO]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " prom");
				pro.add(site);
			}
		}
		return pro;
	}
	
	/**
	 * 
	 * @return	a collection of FeatureSite object which represent
	 * all the Origin features
	 */
	public Vector getOrigin() {

		String temp = "";
		int index = 0;
		int last = 0;
		
		Vector ori = new Vector(5);
		FeatureSite site = null;
		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();

			if (temp.indexOf("[ORI]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " origin");
				ori.add(site);
			}
		}

		return ori;
	}
	
	/**
	 * 
	 * @return a collection of FeatureSite objects which represent
	 * all the regulatory sequence features
	 */
	public Vector getRegulatorySeq() {

		String temp = "";
		int index = 0;
		int last = 0;
		
		Vector reg = new Vector(5);
		FeatureSite site = null;

		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();
			if (temp.indexOf("[REG]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " reg");

				reg.add(site);
			}
		}
		return reg;
	}
	
	/**
	 * 
	 * @return a collection of FeatureSite objects which represent
	 * all the Tag features in the sequence
	 */
	public Vector getTag() {

		String temp = "";
		int index = 0;
		int last = 0;
		Vector tag = new Vector(5);
		FeatureSite site = null;

		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();

			if (temp.indexOf("[TAG]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " tag");
				tag.add(site);
			}
		}
		return tag;
	}
	
	/**
	 * 
	 * @return a collection of FeatureSite objects which represent
	 * all the Terminator features in the sequence
	 */
	public Vector getTerminator() {

		String temp = "";
		int index = 0;
		int last = 0;
		Vector ter = new Vector(5);
		FeatureSite site = null;

		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();

			if (temp.indexOf("[TER]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " term");

				ter.add(site);
			}
		}

		return ter;
	}
	
	/**
	 * 
	 * @return a collection of FeatureSite objects which represent
	 * all the selectable markers in the sequence
	 */
	public Vector getSelector() {

		String temp = "";
		int index = 0;
		int last = 0;
		Vector sel = new Vector(5);
		FeatureSite site = null;

		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();

			if (temp.indexOf("[SEL]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " marker");
				sel.add(site);
			}
		}

		return sel;
	}

	/**
	 * 
	 * @return a collection of FeatuerSite objects which represent
	 * all the reporter genes feature in the sequence
	 */
	public Vector getReporter() {

		String temp = "";
		int index = 0;
		int last = 0;
		Vector rep = new Vector(5);
		FeatureSite site = null;
		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();
			if (temp.indexOf("[REP]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " reporter");
				rep.add(site);
			}
		}

		return rep;
	}
	
	/**
	 * 
	 * @return	a collection of FeatureSite objects which represent
	 * all the locallization feature in the sequence
	 */
	public Vector getLocalization() {

		String temp = "";
		int index = 0;
		int last = 0;
		Vector loc = new Vector(5);
		FeatureSite site = null;

		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();

			if (temp.indexOf("[LOC]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " loc");
				loc.add(site);
			}
		}

		return loc;
	}

	/**
	 * 
	 * @return a collection of FreatureSite objects which represent
	 * all the hybridazation features in the sequence
	 */
	public Vector getHybridazation() {

		String temp = "";
		int index = 0;
		int last = 0;
		Vector hyb = new Vector(5);
		FeatureSite site = null;
		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();
			if (temp.indexOf("[HYB]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last));
				hyb.add(site);
			}
		}
		return hyb;
	}

	/**
	 * 
	 * @return	a collection for FeatureSite objects which represent
	 * all the other features in the sequence
	 */
	public Vector getOther() {

		String temp = "";
		int index = 0;
		int last = 0;
		Vector oth = new Vector(5);
		FeatureSite site = null;
		for (int i = 0; i < hitFeature.size(); i++) {
			site = (FeatureSite) hitFeature.elementAt(i);
			temp = site.getLabel();
			if (temp.indexOf("[OTH]") >= 0) {
				index = temp.indexOf('{');
				last = temp.indexOf('}');
				site.setLabel(temp.substring(index + 1, last) + " other");
				oth.add(site);
			}
		}

		return oth;
	}

	public static void main(String[] args) throws Exception {

		
		
		String[] strands = new String[2];
		strands[0] = "1";
		strands[1] = "2";
		//System.out.println(plasFeature.getORFs(1, strands));

	}
}
