package ca.ualberta.xdong.plasMapper.annotate;
import org.biojava.bio.symbol.*;
import org.biojava.bio.seq.*;
import java.util.*;

/**
 * This class defines some methods to manipulate the plasmid 
 * 
 * @author xdong
 * 
 */

public class PlasmidTool {

	private String dnaSeq;
	private Vector frames;

	public PlasmidTool(String sequence) {

		this.dnaSeq = sequence;
		this.frames = generateFrames();

	}

	/**
	 * 
	 * @return a collection of frames(six)
	 */
	public Vector getFrames() {

		return this.frames;
	}

	/**
	 * 
	 * @return  Get reverse complenet sequence of the given sequence
	 */
	public String getComplement() {

		String dnaCom = "";

		try {
			SymbolList complement = DNATools.createDNA(dnaSeq);

			//complement it
			complement = DNATools.complement(complement);

			dnaCom = complement.seqString();

		} catch (IllegalAlphabetException ex) {

			/*
			* this will occur if you try and transcribe a non DNA sequence or translate
			* a sequence that isn't a triplet view on a RNA sequence.
			*/
			ex.printStackTrace();

		} catch (IllegalSymbolException ex) {
			// this will happen if non IUB characters are used to create the DNA SymbolList
			ex.printStackTrace();
		}

		return dnaCom;
	}

	/**
	 * 
	 * @return Get reverse complenet sequence of the given sequence
	 */

	public String getReverseComplement() {

		String dnaReverse = "";

		try {
			SymbolList reverseComplement = DNATools.createDNA(dnaSeq);

			//reverse complement it
			reverseComplement = DNATools.reverseComplement(reverseComplement);

			dnaReverse = reverseComplement.seqString();

		} catch (IllegalAlphabetException ex) {

			/*
			* this will occur if you try and transcribe a non DNA sequence or translate
			* a sequence that isn't a triplet view on a RNA sequence.
			*/
			ex.printStackTrace();

		} catch (IllegalSymbolException ex) {
			// this will happen if non IUB characters are used to create the DNA SymbolList
			ex.printStackTrace();
		}

		return dnaReverse;
	}

	/**
	 * 
	 * @param frameIndex
	 * @return a translation sequence of the given frame of the sequence
	 */
	public String getTranslation(int frameIndex) {

		String proteinSymL = "";

		try {

			String frame = (String) this.frames.elementAt(frameIndex - 1);

			/*
			  translate the first frame
			*/
			SymbolList symL = DNATools.createDNA(frame);

			//transcribe to RNA
			symL = RNATools.transcribe(symL);

			//translate to protein
			symL = RNATools.translate(symL);
			proteinSymL = symL.seqString();

		} catch (IllegalAlphabetException ex) {

			/*
			 * this will occur if you try and transcribe a non DNA sequence or translate
			 * a sequence that isn't a triplet view on a RNA sequence.
			 */
			ex.printStackTrace();

		} catch (IllegalSymbolException ex) {
			// this will happen if non IUB characters are used to create the DNA SymbolList
			ex.printStackTrace();
		}

		return proteinSymL;
	}

	/**
	 * get all the orfs that is longer than minLen.
	 * <p>
	 * open reading frame is the continus sequence before stop coden
	 * 
	 * @param minLen
	 * @return	a collection of orfs
	 */
	public Vector getORFs(int minLen) {

		Vector orfs = getORFs();

		return filterORFs(orfs, minLen);
	}

	/**
	 * Get all the open reading frame for the sequence
	 * <p>
	 * open reading frame is the continus sequence before stop coden
	 * 
	 * @return a vector of open reading frame. 
	 */
	public Vector getORFs() {

		OpenReadingFrame orf = null;
		Vector orfs = new Vector(10);

		//orf will not include stop coden

		for (int frameIndex = 1; frameIndex <= 6; frameIndex++) {

			String proteinSeq = getTranslation(frameIndex);
			//System.out.println(proteinSeq);

			int length = proteinSeq.length();
			int firstStar = proteinSeq.indexOf('*');
			int lastStar = proteinSeq.lastIndexOf('*');

			//System.out.println("firstStar=" + firstStar + "lastStar=" + lastStar + "length=" + length);
			//there is no stop coden
			if (firstStar == -1) {
				orf =
					new OpenReadingFrame(
						0,
						length - 1,
						proteinSeq.substring(0),
						frameIndex);
				orfs.add(orf);
			}

			//there are multiple stop coden
			else {

				int starIndex = firstStar;

				//the first letter is not stop coden
				if (starIndex != 0) {
					orf =
						new OpenReadingFrame(
							0,
							firstStar,
							proteinSeq.substring(0, firstStar),
							frameIndex);
					orfs.add(orf);
				}
				while (starIndex <= lastStar) {

					//System.out.println("lastStar=" + lastStar);

					if (starIndex == lastStar) {

						if (lastStar == length - 1) {
							break;
						}
						//the stop coden is not the last letter in the sequence
						else {

							orf =
								new OpenReadingFrame(
									starIndex + 1,
									length - 1,
									proteinSeq.substring(
										starIndex + 1,
										length - 1),
									frameIndex);
							orfs.add(orf);
							break;
						}
					}

					int stopIndex = proteinSeq.indexOf("*", starIndex + 1);

					if (stopIndex == -1) {

						orf =
							new OpenReadingFrame(
								starIndex,
								length - 1,
								proteinSeq.substring(starIndex, length - 1),
								frameIndex);
						orfs.add(orf);
						break;
					}

					orf =
						new OpenReadingFrame(
							starIndex + 1,
							stopIndex,
							proteinSeq.substring(starIndex + 1, stopIndex),
							frameIndex);
					orfs.add(orf);

					starIndex = stopIndex;

				}
			}

		}
		return orfs;
	}

	/**
	 * get all the orfs that is longer than minLen.
	 * <p>
	 * open reading frame is the continus sequence 
	 * between stop coden and start codon
	 * 
	 * @param minLen
	 * @return	a collection of orfs
	 */
	public Vector getStartORFs(int minLen) {

		Vector orfs = getStartORFs();

		return filterORFs(orfs, minLen);
	}

	/**
	 * Get all the open reading frame for the sequence
	 * <p>
	 * open reading frame is the continus sequence between stop 
	 * codon and start codon
	 * 
	 * @return a vector of open reading frame. 
	 */
	public Vector getStartORFs() {

		OpenReadingFrame orf = null;
		Vector orfs = new Vector(10);

		//orf will not include stop coden

		for (int frameIndex = 1; frameIndex <= 6; frameIndex++) {

			String proteinSeq = getTranslation(frameIndex);
			//System.out.println(proteinSeq);
			int length = proteinSeq.length();
			int firstStar = proteinSeq.indexOf('*');
			int lastStar = proteinSeq.lastIndexOf('*');

			/*
			there is no stop codon and ignore the sequence
			
			*/
			if (firstStar == -1) {
				int m = proteinSeq.indexOf("M");
				if (m >= 0)
					orf =
						new OpenReadingFrame(
							m,
							length - 1,
							proteinSeq.substring(m),
							frameIndex);
				else
					orf =
						new OpenReadingFrame(
							0,
							length - 1,
							proteinSeq,
							frameIndex);

				orfs.add(orf);
			}

			//there are multiple stop coden
			else {

				int starIndex = firstStar;

				//the first letter is not stop coden
				if (starIndex != 0) {

					String subSeq = proteinSeq.substring(0, firstStar);
					int m = subSeq.indexOf("M");

					if (m >= 0) {
						orf =
							new OpenReadingFrame(
								m,
								firstStar,
								proteinSeq.substring(m, firstStar),
								frameIndex);
						orfs.add(orf);
					}
				}

				while (starIndex <= lastStar) {

					if (starIndex == lastStar) {

						if (lastStar == length - 1) {
							break;
						}
						//the stop coden is not the last letter in the sequence
						else {

							String subSeq =
								proteinSeq.substring(starIndex + 1, length - 1);

							int m = subSeq.indexOf("M");
							if (m >= 0) {
								orf =
									new OpenReadingFrame(
										starIndex + 1 + m,
										length - 1,
										proteinSeq.substring(
											starIndex + 1 + m,
											length - 1),
										frameIndex);
								orfs.add(orf);
							}
							break;
						}
					}

					int stopIndex = proteinSeq.indexOf("*", starIndex + 1);

					if (stopIndex == -1) {
						break;
					}

					String subSeq =
						proteinSeq.substring(starIndex + 1, stopIndex);
					int m = subSeq.indexOf("M");
					if (m >= 0) {
						orf =
							new OpenReadingFrame(
								starIndex + 1 + m,
								stopIndex,
								proteinSeq.substring(
									starIndex + 1 + m,
									stopIndex),
								frameIndex);
						orfs.add(orf);
					}

					starIndex = stopIndex;

				}
			}

		}
		return orfs;
	}

	public Vector filterORFs(Vector orfs, int minLen) {

		Vector targetORFs = new Vector(10);
		for (int i = 0; i < orfs.size(); i++) {

			OpenReadingFrame orf = (OpenReadingFrame) orfs.elementAt(i);

			if (orf.getLength() >= minLen)
				targetORFs.add(orf);
		}

		return targetORFs;

	}

	public Vector sortORFs(Vector orfs) {

		for (int i = 0; i < (orfs.size() - 1); i++) {

			Comparable c = (Comparable) orfs.elementAt(i);

			for (int j = i + 1; j < orfs.size(); j++) {

				if (c.compareTo(orfs.elementAt(j)) > 0)
					c = (Comparable) orfs.elementAt(j);
			}
			if (c != orfs.elementAt(i)) {
				// remove new lowest element and insert at i
				orfs.remove(c);
				orfs.insertElementAt(c, i);
			}
		}

		return orfs;

	}


	/**
	 * Get text output. Mark features in the location in the 
	 * right place of the textoutput
	 * 
	 * @param locations
	 * @return
	 */
	public String getSeqTextOutput(Vector locations) {

		String output = "";
		int lastLocation = 1;
		boolean bar = true;
		boolean moreLabel = false;
		String newLine = "";
		String newSeqLine = "";

		PlasmidTool pt = new PlasmidTool(this.dnaSeq);
		String complement = pt.getComplement();
		String comLine = "";
		//whole sequence control
		for (int seqIndex = 0;
			seqIndex < this.dnaSeq.length();
			seqIndex = seqIndex + 60) {

			int seqStartNum = seqIndex + 1;
			String startStr = "";
			int end = 1;

			if (seqIndex < 10)
				startStr = seqStartNum + "     ";
			else if (seqIndex < 100)
				startStr = seqStartNum + "    ";
			else if (seqIndex < 1000)
				startStr = seqStartNum + "   ";
			else if (seqIndex < 10000)
				startStr = seqStartNum + "  ";
			else if (seqIndex > 10000)
				startStr = seqStartNum + " ";

			bar = true;

			int endLine = seqIndex + 60;

			//output new line of the Seq
			if (endLine < this.dnaSeq.length()) {
				newLine = this.dnaSeq.substring(seqIndex, endLine);
				comLine =
					(complement.substring(seqIndex, endLine)).toUpperCase();
				end = seqIndex + 60;
			} else {
				newLine = this.dnaSeq.substring(seqIndex);
				comLine = (complement.substring(seqIndex)).toUpperCase();
				end = this.dnaSeq.length();
			}

			newSeqLine = "";

			Vector locationsInLine =
				getAllLocationInLine(locations, seqIndex, seqIndex + 60);

			int allLabelNum = locationsInLine.size();

			if (allLabelNum > 0)
				moreLabel = true;

			//ready to add | and labels
			for (int i = 0; i < allLabelNum; i++) {

				Vector labelLeft = new Vector(5);
				//add |
				if (moreLabel == true && bar == true) {

					//lastLocation = 0;
					lastLocation = 1;
					int lastStart = -1;

					String barStr = "";

					for (int k = 0; k < locationsInLine.size(); k++) {

						FeatureSite site =
							(FeatureSite) locationsInLine.elementAt(k);

						//int start = loc.getStart();
						int start = site.getStart();

						if (start >= 60)
							start = start % 60;

						start = start + 6;

						if (start > lastStart) {
							for (int j = lastLocation; j < start; j++) {

								//output = output + " ";

								barStr = barStr + " ";
							}
							lastStart = start;
							lastLocation = start + 1;
							//output = output + "|";
							barStr = barStr + "|";

						}
					}
					//output = output + "\n";
					newSeqLine = barStr + "\n" + newSeqLine;

					bar = false;
				}

				//add labels
				if (moreLabel == true && bar == false) {

					//int lastLabelLoc = 0;
					int lastLabelLoc = 1;
					int k = 0;

					String labelsLine = "";
					for (; k < locationsInLine.size(); k++) {

						FeatureSite site =
							(FeatureSite) locationsInLine.elementAt(k);

						//int labelStart = site.getStart() + 1;
						int labelStart = site.getStart();
						String label = "";
						System.err.println("start=" + labelStart + " END=" + site.getEnd());
						if (labelStart == site.getEnd())
							label = site.getLabel();
						else {

							if ((site.getDirection()).equals("1"))
								label =
									site.getLabel()
										+ "("
										+ labelStart
										+ ","
										+ site.getEnd()
										+ ")>>>";

							else if ((site.getDirection()).equals("2"))
								label =
									site.getLabel()
										+ "("
										+ labelStart
										+ ","
										+ site.getEnd()
										+ ")<<<";
							
						}

						int start = site.getStart();
						if (start >= 60)
							start = start % 60;
						start = start + 6;

						//start location has been occupied

						if (lastLabelLoc > start) {
							labelLeft.add(site);

						} else {

							for (int j = lastLabelLoc; j < start; j++) {

								//output = output + " ";
								labelsLine = labelsLine + " ";
							}

							lastLabelLoc = start + label.length() + 1;

							//output = output + label + " ";
							labelsLine = labelsLine + label + " ";
						}
					}

					if (labelLeft.size() > 0) {

						moreLabel = true;
						bar = true;
						locationsInLine = labelLeft;

					} else {

						moreLabel = false;

					}

					//output = output + "\n";
					newSeqLine = labelsLine + "\n" + newSeqLine;

				}

			}
			newSeqLine =
				newSeqLine
					+ startStr
					+ newLine
					+ " "
					+ end
					+ "\n"
					+ "      "
					+ comLine
					+ "\n";

			output = output + newSeqLine + "\n";
		}
		return output;
		//System.out.println(output);
	}

	/**
	 * 
	 * @return get all six frames and put then into the vector
	 */
	private Vector generateFrames() {

		String dnaSeq1 = "";
		String dnaSeq2 = "";
		String dnaSeq3 = "";
		String dnaSeq4 = "";
		String dnaSeq5 = "";
		String dnaSeq6 = "";

		int length = this.dnaSeq.length();
		Vector frames = new Vector(6);

		if (dnaSeq.length() % 3 == 0) {
			dnaSeq1 = dnaSeq;
			dnaSeq2 = dnaSeq.substring(1, length - 2);
			dnaSeq3 = dnaSeq.substring(2, length - 1);
		} else if (dnaSeq.length() % 3 == 1) {
			dnaSeq1 = dnaSeq.substring(0, length - 1);
			dnaSeq2 = dnaSeq.substring(1);
			dnaSeq3 = dnaSeq.substring(2, length - 2);
		} else if (dnaSeq.length() % 3 == 2) {
			dnaSeq1 = dnaSeq.substring(0, length - 2);
			dnaSeq2 = dnaSeq.substring(1, length - 1);
			dnaSeq3 = dnaSeq.substring(2);
		}

		String dnaReverse = getReverseComplement();
		//String dnaReverse = getComplement();
		if (dnaReverse.length() % 3 == 0) {
			dnaSeq4 = dnaReverse;
			dnaSeq5 = dnaReverse.substring(1, length - 2);
			dnaSeq6 = dnaReverse.substring(2, length - 1);
		} else if (dnaReverse.length() % 3 == 1) {
			dnaSeq4 = dnaReverse.substring(0, length - 1);
			dnaSeq5 = dnaReverse.substring(1);
			dnaSeq6 = dnaReverse.substring(2, length - 2);
		} else if (dnaReverse.length() % 3 == 2) {
			dnaSeq4 = dnaReverse.substring(0, length - 2);
			dnaSeq5 = dnaReverse.substring(1, length - 1);
			dnaSeq6 = dnaReverse.substring(2);
		}

		frames.add(dnaSeq1);
		frames.add(dnaSeq2);
		frames.add(dnaSeq3);
		frames.add(dnaSeq4);
		frames.add(dnaSeq5);
		frames.add(dnaSeq6);
		return frames;

	}

	/**
	 * Get all the FeatureSite in this line of the sequence
	 * 
	 * @param locations a vector the FeatureSite objects
	 * @param start	start postion of the line
	 * @param end	end position of the line
	 * @return all the featureSite objects in this line
	 */
	public static Vector getAllLocationInLine(
		Vector locations,
		int start,
		int end) {

		Vector vector = new Vector(10);
		for (int i = 0; i < locations.size(); i++) {

			FeatureSite location = (FeatureSite) locations.elementAt(i);

			if (location.getStart() >= end)
				break;
			else if (location.getStart() >= start && location.getStart() < end)
				vector.add(location);
		}

		return vector;
	}

	public static void main(String[] args) {

		String seq = "ATGAAAACTGCTAAAACTGCTTAAGCTTCGATGCAAATGTAAGCAATG";

		StringBuffer buffer = new StringBuffer(1000);
		StringTokenizer st = new StringTokenizer(seq);
		while (st.hasMoreTokens()) {
			buffer.append(st.nextToken());
		}
		System.out.println(seq);
		String dnaSeq = buffer.toString();
		PlasmidTool tra = new PlasmidTool(dnaSeq);
		System.out.println(tra.getComplement());
		System.out.println(tra.getReverseComplement());

		String tem = tra.getTranslation(1);
		System.out.println("frame1");
		System.out.println(tem);

		tem = tra.getTranslation(2);
		System.out.println("frame2");
		System.out.println(tem);

		tem = tra.getTranslation(3);
		System.out.println("frame3");
		System.out.println(tem);

		tem = tra.getTranslation(4);
		System.out.println("frame4");
		System.out.println(tem);

		tem = tra.getTranslation(5);
		System.out.println("frame5");
		System.out.println(tem);

		tem = tra.getTranslation(6);
		System.out.println("frame6");
		System.out.println(tem);

		Vector orfs = tra.getORFs();
		System.out.println("all ORFs");
		System.out.println(orfs);

		orfs = tra.getStartORFs();
		System.out.println("start ORFs");
		System.out.println(orfs);

		Vector targetORFs = tra.filterORFs(orfs, 1);
		System.out.println("ORFs after filter");
		System.out.println(targetORFs);

		System.out.println("After sorting");
		System.out.println(tra.sortORFs(orfs));

		seq = "ATGAAAACTGCTAAAACTGCTTAAGCTTCGATGCAAATGTAAGCAATG";

		Vector locations = new Vector(5);
		FeatureSite location = new FeatureSite(1, 5, "xiaoli", "1");
		locations.add(location);

		location = new FeatureSite(7, 7, "Haiyan", "0");
		locations.add(location);

		location = new FeatureSite(7, 15, "nelson", "1");
		locations.add(location);

		location = new FeatureSite(7, 15, "maodou", "1");
		locations.add(location);

		location = new FeatureSite(17, 25, "paul", "1");
		locations.add(location);
		
		location = new FeatureSite(28, 28, "paul", "0");
		locations.add(location);

		location = new FeatureSite(59, 25, "anchi", "1");
		locations.add(location);

		location = new FeatureSite(60, 25, "raj", "1");
		locations.add(location);

		location =
			new FeatureSite(
				62,
				25,
				"rajkkkkkkkkkkkkkkkkkkkkkknnnn_dslfkdasl;fka'sdf",
				"1");
		locations.add(location);

		location = new FeatureSite(65, 25, "trent", "1");
		locations.add(location);

		locations = FeatureSite.bubbleSort(locations, locations.size());
		PlasmidTool tool = new PlasmidTool(seq);
		String textMap = tool.getSeqTextOutput(locations);

		System.out.println(textMap);

	}
}
